@@ -1,4 +1,4 @@
-/* $Id: input.c,v 1.7 2002/06/08 19:51:06 skids Exp $
+/* $Id: input.c,v 1.17 2004/09/16 13:25:01 pekberg Exp $
 ******************************************************************************
 
    Xwin inputlib - use an existing X window as a LibGII input source
@@ -32,6 +32,7 @@
 #include "config.h"
 #include <ggi/internal/gii-dl.h>
 #include <ggi/internal/gii_debug.h>
+#include <ggi/internal/gii_target.h>
 #include <ggi/input/xwin.h>
 #include "../x/xev.h"
 
@@ -49,6 +50,7 @@ enum {
 typedef struct {
 	Display	       *disp;
 	Window		win;
+	Window		parentwin;
 	XComposeStatus	compose_status;
 	XIM		xim;
 	XIC		xic;
@@ -67,6 +69,7 @@ typedef struct {
 	void	*lock;
 
 	uint32   origin[XWIN_DEV_MAX];
+	char	 key_vector[32];
 } xwin_priv;
 
 #define XWIN_PRIV(inp)	((xwin_priv*)((inp)->priv))
@@ -139,13 +142,13 @@ static inline void update_winparam(xwin_priv *priv)
 {
 	if (!priv->alwaysrel) {
 		Window dummywin;
-		unsigned int w, h;
-		int dummy;
-		
+		unsigned int w, h, dummy;
+
 		if (priv->cursor == None) {
 			priv->cursor = make_cursor(priv->disp, priv->win);
 		}
-		XGetGeometry(priv->disp, priv->win, &dummywin, &dummy, &dummy,
+		XGetGeometry(priv->disp, priv->win, &dummywin,
+			     (int *)&dummy, (int *)&dummy,
 			     &w, &h, &dummy, &dummy);
 		priv->width = w;
 		priv->height = h;
@@ -210,6 +213,8 @@ static gii_event_mask GII_xwin_eventpoll(gii_input *inp, void *arg)
 			giiev.any.origin = priv->origin[XWIN_DEV_KEY];
 			giiev.key.button = keycode - 8;
 
+			priv->key_vector[keycode/8]|=1<<(keycode&7);
+
 			if (havecached &&
 			    releasecache.key.button == giiev.key.button) {
 				if (xev.xkey.time == releasetime) {
@@ -251,6 +256,7 @@ static gii_event_mask GII_xwin_eventpoll(gii_input *inp, void *arg)
 			break;
 
 		case KeyRelease:
+			priv->key_vector[keycode/8]&=~(1<<(keycode&7));
 			if (havecached) {
 				_giiEvQueueAdd(inp, &releasecache);
 				rc |= emKeyRelease;
@@ -285,6 +291,67 @@ static gii_event_mask GII_xwin_eventpoll(gii_input *inp, void *arg)
 			GIIDPRINT_EVENTS("GII_xwin_eventpoll: KeyRelease\n");
 			break;
 
+		case KeymapNotify: {
+			unsigned char delta;
+			int j;
+			XEvent fake_xev;
+			
+			for(i=1;i<32;i++) {
+				delta=priv->key_vector[i]^xev.xkeymap.key_vector[i];
+				if (!delta) continue;
+				for(j=0;j<8;j++) {
+					if (0==(delta&(1<<j))) continue;
+					keycode=i*8+j;
+					fake_xev.xkey.serial	=xev.xkeymap.serial;
+					fake_xev.xkey.send_event=xev.xkeymap.send_event;
+					fake_xev.xkey.display	=xev.xkeymap.display;
+					fake_xev.xkey.window	=xev.xkeymap.window;
+					fake_xev.xkey.root	=xev.xkeymap.window;
+					fake_xev.xkey.subwindow	=xev.xkeymap.window;
+					fake_xev.xkey.time	=0;
+					fake_xev.xkey.x		=0;
+					fake_xev.xkey.y		=0;
+					fake_xev.xkey.x_root	=0;
+					fake_xev.xkey.y_root	=0;
+					fake_xev.xkey.state	=0;
+					fake_xev.xkey.keycode	=keycode;
+					fake_xev.xkey.same_screen=0;
+					
+					if (xev.xkeymap.key_vector[i]&(1<<j)) {
+						giiev.any.size = sizeof(gii_key_event);
+						giiev.any.type = evKeyPress;
+						giiev.any.origin = priv->origin[XWIN_DEV_KEY];
+						giiev.key.button = keycode - 8;
+						rc |= emKeyPress;
+						fake_xev.xkey.type      =KeyPress;
+						_gii_xev_trans(&fake_xev.xkey, &giiev.key,
+								&priv->compose_status, priv->xic,
+								&priv->oldcode);
+						if (giiev.any.type == evKeyPress &&
+						    giiev.key.label < 0x60) {
+							priv->symstat[giiev.key.label] = giiev.key.sym;
+						}
+						_giiEvQueueAdd(inp, &giiev);
+					} else {
+						giiev.any.size = sizeof(gii_key_event);
+						giiev.any.type = evKeyRelease;
+						giiev.any.origin = priv->origin[XWIN_DEV_KEY];
+						giiev.key.button = keycode - 8;
+						rc |= emKeyRelease;
+						fake_xev.xkey.type      =KeyRelease;
+						_gii_xev_trans(&fake_xev.xkey, &giiev.key,
+								&priv->compose_status, NULL, NULL);
+						if (giiev.key.label < 0x60 &&
+						    priv->symstat[giiev.key.label] != 0) {
+							giiev.key.sym=priv->symstat[giiev.key.label];
+						}
+						_giiEvQueueAdd(inp, &giiev);
+					}
+				}
+				priv->key_vector[i]=xev.xkeymap.key_vector[i];
+			}
+			} goto dont_queue_this_event;
+
 		case ButtonPress:
 			giiev.any.type = evPtrButtonPress;
 			giiev.any.origin = priv->origin[XWIN_DEV_MOUSE];
@@ -315,7 +382,8 @@ static gii_event_mask GII_xwin_eventpoll(gii_input *inp, void *arg)
 					if (ABS(priv->width/2 - xev.xmotion.x)
 					    > priv->width / 4 ||
 					    ABS(priv->height/2 - xev.xmotion.y)
-					    > priv->height / 4) {
+					    > priv->height / 4)
+					{
 						center_pointer(priv);
 					}
 #undef ABS			
@@ -345,9 +413,18 @@ static gii_event_mask GII_xwin_eventpoll(gii_input *inp, void *arg)
 			break;
 
 		case EnterNotify:
+			/* We used to pull focus here using
+			 * XSetInputFocus(priv->disp, priv->win,
+			 *		RevertToParent, CurrentTime);
+			 * We now leave that to the window manager.
+			 */
 		case LeaveNotify:
 			break;
 
+		case FocusIn:
+		case FocusOut:
+			break;
+
 		case Expose:
 			GIIDPRINT_EVENTS("GII_xwin_eventpoll: Expose\n");
 			if (priv->exposefunc != NULL) {
@@ -355,7 +432,8 @@ static gii_event_mask GII_xwin_eventpoll(gii_input *inp, void *arg)
 						       xev.xexpose.x,
 						       xev.xexpose.y,
 						       xev.xexpose.width,
-						       xev.xexpose.height)) {
+						       xev.xexpose.height))
+				{
 					goto dont_queue_this_event;
 				}
 			}
@@ -371,13 +449,13 @@ static gii_event_mask GII_xwin_eventpoll(gii_input *inp, void *arg)
 		case GraphicsExpose:
 			GIIDPRINT_EVENTS("GII_xwin_eventpoll: Expose dest\n");
 			if (priv->exposefunc != NULL) {
-				if (!
-				    priv->exposefunc(priv->exposearg,
+				if (! priv->exposefunc(priv->exposearg,
 						     xev.xgraphicsexpose.x,
 						     xev.xgraphicsexpose.y,
 						     xev.xgraphicsexpose.width,
 						     xev.xgraphicsexpose.height
-						     )) {
+						     ))
+				{
 					goto dont_queue_this_event;
 				}
 			}
@@ -479,8 +557,11 @@ static void send_devinfo(gii_input *inp, int dev)
 {
 	gii_event ev;
 	gii_cmddata_getdevinfo *dinfo;
-	int size = sizeof(gii_cmd_nodata_event)+sizeof(gii_cmddata_getdevinfo);
+	size_t size;
 	xwin_priv  *priv = XWIN_PRIV(inp);
+
+	size = sizeof(gii_cmd_nodata_event)+sizeof(gii_cmddata_getdevinfo);
+
 	_giiEventBlank(&ev, size);
 	
 	ev.any.size   = size;
@@ -504,11 +585,11 @@ static void send_devinfo(gii_input *inp, int dev)
 }
 
 
-static int
-GIIsendevent(gii_input *inp, gii_event *ev)
+static int GIIsendevent(gii_input *inp, gii_event *ev)
 {
 	if ((ev->any.target & 0xffffff00) != inp->origin &&
-	    ev->any.target != GII_EV_TARGET_ALL) {
+	    ev->any.target != GII_EV_TARGET_ALL)
+	{
 		/* not for us */
 		return GGI_EEVNOTARGET;
 	}
@@ -534,14 +615,21 @@ GIIsendevent(gii_input *inp, gii_event *ev)
 			}
 			return GGI_EEVNOTARGET;
 		}
-		
+
 	} else if (ev->cmd.code == GII_CMDCODE_XWINSETPARAM) {
 		xwin_priv  *priv = XWIN_PRIV(inp);
-		gii_xwin_cmddata_setparam *data	= (void *) ev->cmd.data;
+		gii_xwin_cmddata_setparam data;
+
+		/* Assure aligned memory access. Some platforms
+		 * (i.e. NetBSD/sparc64) rely on this.
+		 */
+		memcpy(&data, (gii_xwin_cmddata_setparam *)ev->cmd.data,
+			sizeof(gii_xwin_cmddata_setparam));
+
+		priv->win = data.win;
+		priv->parentwin = data.parentwin;
+		priv->alwaysrel = data.ptralwaysrel;
 
-		priv->win = data->win;
-		priv->alwaysrel = data->ptralwaysrel;
-		
 		update_winparam(priv);
 
 		return 0;
@@ -571,13 +659,19 @@ int GIIdlinit(gii_input *inp, const char *args, void *argptr)
 	if (xwinarg == NULL || xwinarg->disp == NULL) {
 		return GGI_EARGREQ;
 	}
-	
+
 	if ((priv = malloc(sizeof(xwin_priv))) == NULL) {
 		return GGI_ENOMEM;
 	}
-	
+
+
+	/* Attention: priv->win is NOT valid at this
+	 * point. It becomes valid once a mode has
+	 * been set up by libggi's X-target.
+	 */
 	priv->disp = xwinarg->disp;
 	priv->win  = xwinarg->win;
+	priv->parentwin  = xwinarg->win;
 	priv->compose_status = dummyxcs;
 	priv->xim = NULL;
 	priv->xic = NULL;
@@ -593,34 +687,34 @@ int GIIdlinit(gii_input *inp, const char *args, void *argptr)
 	priv->resizefunc	= xwinarg->resizefunc;
 	priv->resizearg		= xwinarg->resizearg;
 	priv->lock		= xwinarg->gglock;
-	
+
+	memset(priv->key_vector,0,sizeof(priv->key_vector));
+
 	if (!xwinarg->wait) {
 		update_winparam(priv);
 	} else {
 		priv->cursor = None;
 	}
-	
+
 	inp->priv = priv;
 
 	inp->GIIsendevent = GIIsendevent;
 	inp->GIIeventpoll = GII_xwin_eventpoll;
 	inp->GIIclose = GII_xwin_close;
-	
-	if(0==(priv->origin[XWIN_DEV_KEY]=
-	       _giiRegisterDevice(inp,&key_devinfo,NULL))) {
-		
-		GII_xwin_close(inp);
-		return GGI_ENOMEM;
-		
-	}
-	
-	if(0==(priv->origin[XWIN_DEV_MOUSE]=
-	       _giiRegisterDevice(inp,&mouse_devinfo,NULL))) {
-		
+
+	if (0 == (priv->origin[XWIN_DEV_KEY] =
+		_giiRegisterDevice(inp,&key_devinfo,NULL)))
+	{
 		GII_xwin_close(inp);
 		return GGI_ENOMEM;
 	}
 
+	if (0 == (priv->origin[XWIN_DEV_MOUSE] =
+		_giiRegisterDevice(inp,&mouse_devinfo,NULL)))
+	{
+		GII_xwin_close(inp);
+		return GGI_ENOMEM;
+	}
 
 
 	inp->targetcan = emKey | emPointer | emExpose;
@@ -632,7 +726,7 @@ int GIIdlinit(gii_input *inp, const char *args, void *argptr)
 	mouse_devinfo.num_buttons = XGetPointerMapping(priv->disp, NULL, 0);
 	XDisplayKeycodes(priv->disp, &minkey, &maxkey);
 	key_devinfo.num_buttons = (maxkey - minkey) + 1;
-	
+
 	send_devinfo(inp, XWIN_DEV_KEY);
 	send_devinfo(inp, XWIN_DEV_MOUSE);
 
